// メモリ節約型設定ファイル読み込みライブラリ (diminutive configuration file loader)

/*!
 * @file dimiConfigLoader.c
 * @brief メモリ節約型ファイル読み込みライブラリのソース
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "dimiConfigLoader.h"

////////////////////////////////
// 定数定義

#define BUFMAX (16*1024)
#define STRMAX   256
#define ENTRYMAX 256

#define GET_NEXT(p,i)      (((*((p)+(i)))>>24)&255)
#define GET_NAME(p,i)      (((*((p)+(i)))>>12)&4095)
#define GET_VALUE(p,i)     (((*((p)+(i)))    )&4095)
#define GET_NAME_PTR(p,i)  ((char *)((p)+GET_NAME(p,i)))
#define GET_VALUE_PTR(p,i) ((char *)((p)+GET_VALUE(p,i)))

#define SET_NEXT(p,i,v)    ((*((p)+(i))) = (((*((p)+(i)))&0x00ffffff)|((v)<<24)))
#define SET_NAME(p,i,v)    ((*((p)+(i))) = (((*((p)+(i)))&0xff000fff)|((v)<<12)))
#define SET_VALUE(p,i,v)   ((*((p)+(i))) = (((*((p)+(i)))&0xfffff000)|((v)    )))
#define SET_ENTRY(p,i,n,v) ((*((p)+(i))) = ((n)<<12)|(v))

////////////////////////////////
// 構造体定義

typedef unsigned long entry_t;

typedef struct {
	entry_t *buf;
	int      pos_free;
	int      pos_text;
} DIMICONF_HANDLE;

////////////////////////////////
// 内部関数定義

static int DimiConf_Load(DIMICONF_HANDLE *handle, const char *fname);
static int DimiConf_SearchSection(DIMICONF_HANDLE *handle, const char *section_name);
static int DimiConf_SearchKey(DIMICONF_HANDLE *handle, int section, const char *key_name);
static int DimiConf_SearchNextKey(DIMICONF_HANDLE *handle, int prev_key, const char *key_name);
static int DimiConf_AppendSection(DIMICONF_HANDLE *handle, const char *section_name);
static int DimiConf_AppendKey(DIMICONF_HANDLE *handle, int section, const char *key_name, const char *value);
static int DimiConf_StrEqu(const char *s1, const char *s2);

////////////////////////////////
// API関数

int DimiConf_Open(void **conf, const char *fname) {
	DIMICONF_HANDLE *handle = NULL;
	int r;

	// 引数チェック
	if (conf == NULL) return DIMICONF_E_PARAM;
	if (fname == NULL) return DIMICONF_E_PARAM;

	// メモリの確保
	handle = (DIMICONF_HANDLE *)malloc(sizeof(DIMICONF_HANDLE)+BUFMAX);
	if (handle == NULL) return DIMICONF_E_MEMORY;

	// ハンドルの初期化
	handle->buf      = (entry_t *)(handle+1);
	handle->pos_free = 1;
	handle->pos_text = BUFMAX/sizeof(entry_t)-1;

	// 名前なしセクションの登録
	SET_ENTRY(handle->buf, 0, handle->pos_text, 0);
	*(GET_NAME_PTR(handle->buf, 0)) = '\0';

	// 設定ファイルの読み込み
	r = DimiConf_Load(handle, fname);

	*conf = (void *)handle;
	return r;
}

int DimiConf_Close(void *conf) {

	// メモリの解放
	free(conf);

	return DIMICONF_OK;
}

char * DimiConf_GetString(void *conf, const char *section_name, const char *key_name, char *def) {
	DIMICONF_HANDLE *handle = (DIMICONF_HANDLE *)conf;
	int section;
	int key;

	// 引数チェック
	if (conf == NULL) return NULL;
	if (section_name == NULL) return NULL;
	if (key_name == NULL) return NULL;

	// セクションの探索
	section = DimiConf_SearchSection(handle, section_name);
	if (section < 0) return def;

	// キーの探索
	key = DimiConf_SearchKey(handle, section, key_name);
	if (key < 0) return def;

	return GET_VALUE_PTR(handle->buf, key);
}

char * DimiConf_GetSiblingString(void *conf, const char *section_name, const char *key_name, int pos, char *def) {
	DIMICONF_HANDLE *handle = (DIMICONF_HANDLE *)conf;
	int section;
	int key;

	// 引数チェック
	if (conf == NULL) return NULL;
	if (section_name == NULL) return NULL;
	if (key_name == NULL) return NULL;
	if (pos < 0) return NULL;

	// セクションの探索
	section = DimiConf_SearchSection(handle, section_name);
	if (section < 0) return def;

	// キーの探索
	key = DimiConf_SearchKey(handle, section, key_name);
	if (key < 0) return def;

	while (pos > 0) {
		--pos;
		key = DimiConf_SearchNextKey(handle, key, key_name);
		if (key < 0) return def;
	}

	return GET_VALUE_PTR(handle->buf, key);
}

////////////////////////////////
// 内部関数

// 設定ファイルの読み込み
static int DimiConf_Load(DIMICONF_HANDLE *handle, const char *fname) {
	FILE *fp = NULL;
	char str[STRMAX];
	char *section_name, *key_name, *value;
	unsigned char *p, *q;
	int   section, key;
	int r = DIMICONF_OK;

	// 名前なしセクションを取得
	section = DimiConf_SearchSection(handle, "");

	// ファイルを開く
	fp = fopen(fname, "r");
	if (fp == NULL) {
		r = DIMICONF_E_OPEN;
		goto fail;
	}

	while (!feof(fp)) {
		// １行読み込む
		p = (unsigned char *)fgets(str, sizeof(str), fp);
		if (p == NULL) break;

		while (isspace(*p)) p++;

		if ((*p == '\0') || (*p == '#')) {
			// 空行・コメント
			;
		} else if (*p == '[') {
			// セクション
			p++;
			while (isspace(*p)) p++;
			section_name = (char *)p;
			while (isalnum(*p) || (*p == '_') || (*p == '.')) p++;
			q = p;
			while (isspace(*p)) p++;
			if (*p == ']') {
				*q = '\0';

				// セクションを検索
				section = DimiConf_SearchSection(handle, section_name);
				if (section < 0) {
					// 未登録であれば新規に追加する
					section = DimiConf_AppendSection(handle, section_name);
					if (section < 0) {
						// バッファ不足エラー
						r = DIMICONF_E_OVERFLOW;
						goto fail;
					}
				}
			} else {
				// 構文エラー
				r = DIMICONF_E_SYNTAX;
				goto fail;
			}
		} else if (isalnum(*p) || (*p == '_') || (*p == '.')) {
			// 項目
			key_name = (char *)p;
			while (isalnum(*p) || (*p == '_') || (*p == '.')) p++;
			q = p;
			while (isspace(*p)) p++;
			if (*p == '=') {
				*q = '\0';
				p++;
				while (isspace(*p)) p++;
				if (*p == '\"') {
					p++;
					value = (char *)p;
					while (*p && (*p != '\"')) p++;
					*p = '\0';
				} else {
					value = (char *)p;
					while (*p && !isspace(*p)) p++;
					*p = '\0';
				}

				// キーを新規に追加する
				key = DimiConf_AppendKey(handle, section, key_name, value);
				if (key < 0) {
					// バッファ不足エラー
					r = DIMICONF_E_OVERFLOW;
					goto fail;
				}
			} else {
				// 構文誤り
				r = DIMICONF_E_SYNTAX;
				goto fail;
			}
		} else {
			// 構文誤り
			r = DIMICONF_E_SYNTAX;
			goto fail;
		}
	}

	r = DIMICONF_OK;
fail:
	// ファイルを閉じる
	if (fp != NULL) fclose(fp);

	return r;
}

// セクションの探索
static int DimiConf_SearchSection(DIMICONF_HANDLE *handle, const char *section_name) {
	entry_t *buf = handle->buf;
	int pos = 0;

	// セクションのリストを順にたどる
	// 最初のエントリーは名無しセクションとして登録済みのため
	// ループの最初でposが終端であるか判断する必要はなし
	do {
		if (DimiConf_StrEqu(GET_NAME_PTR(buf,pos), section_name)) return pos;
		pos = GET_NEXT(buf, pos);
	} while (pos != 0);

	return -1;
}

// キーの探索
static int DimiConf_SearchKey(DIMICONF_HANDLE *handle, int section, const char *key_name) {
	entry_t *buf = handle->buf;
	int pos = GET_VALUE(buf, section);

	// キーのリストを順にたどる
	// セクションにキーが未登録の場合もあるので、
	// ループの最初でposが終端か判断する必要あり
	while (pos != 0) {
		if (DimiConf_StrEqu(GET_NAME_PTR(buf,pos), key_name)) return pos;
		pos = GET_NEXT(buf, pos);
	}

	return -1;
}

// 次のキーの探索
static int DimiConf_SearchNextKey(DIMICONF_HANDLE *handle, int prev_key, const char *key_name) {
	entry_t *buf = handle->buf;
	int pos = GET_NEXT(buf, prev_key);

	// キーのリストを順にたどる
	// セクションにキーが未登録の場合もあるので、
	// ループの最初でposが終端か判断する必要あり
	while (pos != 0) {
		if (DimiConf_StrEqu(GET_NAME_PTR(buf,pos), key_name)) return pos;
		pos = GET_NEXT(buf, pos);
	}

	return -1;
}

// セクションの追加
static int DimiConf_AppendSection(DIMICONF_HANDLE *handle, const char *section_name) {
	entry_t *buf = handle->buf;
	int   pos;
	int   npos, nlen;

	// 登録済みエントリー数のチェック
	if (handle->pos_free >= ENTRYMAX) return -1;

	// セクション名を格納する領域のサイズを求める
	nlen = ((int)strlen(section_name)+1+sizeof(entry_t)-1) / sizeof(entry_t);

	// 必要なサイズの未使用領域があるか判定する
	if (handle->pos_free+1 > handle->pos_text-nlen) return -1;

	// セクションの末尾のエントリーへ移動し、新規エントリーへのリンクを張る
	pos = 0;
	while (GET_NEXT(buf, pos) != 0) {
		pos = GET_NEXT(buf, pos);
	}
	SET_NEXT(buf, pos, handle->pos_free);

	// セクション名を格納する
	npos = (handle->pos_text -= nlen);
	strcpy((char *)(buf+npos), section_name);

	// 新規エントリーを追加する
	pos = (handle->pos_free)++;
	SET_ENTRY(buf, pos, npos, 0);

	return pos;
}

// キーの追加
static int DimiConf_AppendKey(DIMICONF_HANDLE *handle, int section, const char *key_name, const char *value) {
	entry_t *buf = handle->buf;
	int pos;
	int npos, nlen;
	int vpos, vlen;

	// 登録済みエントリー数のチェック
	if (handle->pos_free >= ENTRYMAX) return -1;

	// キー名と設定値を格納する領域のサイズを求める
	nlen = ((int)strlen(key_name)+1+sizeof(entry_t)-1) / sizeof(entry_t);
	vlen = ((int)strlen(value)   +1+sizeof(entry_t)-1) / sizeof(entry_t);

	// 必要なサイズの未使用領域があるか判定する
	if (handle->pos_free+1 > handle->pos_text-nlen-vlen) return -1;

	// 指定したセクションのキーの末尾のエントリーへ移動し、
	// 新規エントリーへのリンクを張る
	pos = GET_VALUE(buf, section);
	if (pos == 0) {
		SET_VALUE(buf, section, handle->pos_free);
	} else {
		while (GET_NEXT(buf, pos) != 0) {
			pos = GET_NEXT(buf, pos);
		}
		SET_NEXT(buf, pos, handle->pos_free);
	}

	// キー名と設定値を格納する
	npos = (handle->pos_text -= nlen);
	vpos = (handle->pos_text -= vlen);
	strcpy((char *)(buf+npos), key_name);
	strcpy((char *)(buf+vpos), value);

	// 新規エントリーを追加する
	pos = (handle->pos_free++);
	SET_ENTRY(buf, pos, npos, vpos);

	return pos;
}

// 文字列の比較
static int DimiConf_StrEqu(const char *s1, const char *s2) {

	// 大文字小文字の区別なく文字列が完全一致するか判定
	while (*s1) {
		if (toupper(*(s1++)) != toupper(*(s2++))) return 0;
	}

	return (*s1 == *s2) ? 1 : 0;
}

